import asyncio
import contextlib
import logging
import sys
import types
from collections.abc import Callable
from logging.config import dictConfig
from multiprocessing import Manager
from pathlib import Path
from threading import Thread

import panther.logging
from panther import status
from panther._load_configs import *
from panther._utils import clean_traceback_message, http_response, is_function_async, reformat_code, \
    check_class_type_endpoint, check_function_type_endpoint
from panther.background_tasks import background_tasks
from panther.cli.utils import print_info
from panther.configs import config
from panther.exceptions import APIException, PantherException
from panther.monitoring import Monitoring
from panther.request import Request
from panther.response import Response
from panther.routings import collect_path_variables, find_endpoint

dictConfig(panther.logging.LOGGING)
logger = logging.getLogger('panther')


class Panther:
    def __init__(self, name: str, configs=None, urls: dict | None = None, startup: Callable = None, shutdown: Callable = None):
        self._configs_module_name = configs
        self._urls = urls
        self._startup = startup
        self._shutdown = shutdown

        config['base_dir'] = Path(name).resolve().parent

        try:
            self.load_configs()
            if config['auto_reformat']:
                reformat_code(base_dir=config['base_dir'])
        except Exception as e:  # noqa: BLE001
            if isinstance(e, PantherException):
                logger.error(e.args[0])
            else:
                logger.error(clean_traceback_message(e))
            sys.exit()

        # Monitoring
        self.monitoring = Monitoring(is_active=config['monitoring'])

        # Print Info
        print_info(config)

    def load_configs(self) -> None:

        # Check & Read The Configs File
        self._configs_module = load_configs_module(self._configs_module_name)

        # Put Variables In "config" (Careful about the ordering)
        config['secret_key'] = load_secret_key(self._configs_module)
        config['monitoring'] = load_monitoring(self._configs_module)
        config['log_queries'] = load_log_queries(self._configs_module)
        config['background_tasks'] = load_background_tasks(self._configs_module)
        config['throttling'] = load_throttling(self._configs_module)
        config['default_cache_exp'] = load_default_cache_exp(self._configs_module)
        config['pantherdb_encryption'] = load_pantherdb_encryption(self._configs_module)
        middlewares = load_middlewares(self._configs_module)
        config['http_middlewares'] = middlewares['http']
        config['ws_middlewares'] = middlewares['ws']
        config['reversed_http_middlewares'] = middlewares['http'][::-1]
        config['reversed_ws_middlewares'] = middlewares['ws'][::-1]
        config['user_model'] = load_user_model(self._configs_module)
        config['authentication'] = load_authentication_class(self._configs_module)
        config['jwt_config'] = load_jwt_config(self._configs_module)
        config['startup'] = load_startup(self._configs_module)
        config['shutdown'] = load_shutdown(self._configs_module)
        config['auto_reformat'] = load_auto_reformat(self._configs_module)
        config['models'] = collect_all_models()

        # Initialize Background Tasks
        if config['background_tasks']:
            background_tasks.initialize()

        # Load URLs should be one of the last calls in load_configs,
        #   because it will read all files and loads them.
        config['flat_urls'], config['urls'] = load_urls(self._configs_module, urls=self._urls)
        config['urls']['_panel'] = load_panel_urls()

        self._create_ws_connections_instance()

    def _create_ws_connections_instance(self):
        from panther.base_websocket import Websocket, WebsocketConnections

        # Check do we have ws endpoint
        for endpoint in config['flat_urls'].values():
            if not isinstance(endpoint, types.FunctionType) and issubclass(endpoint, Websocket):
                config['has_ws'] = True
                break
        else:
            config['has_ws'] = False

        # Create websocket connections instance
        if config['has_ws']:
            # Websocket Redis Connection
            for middleware in config['http_middlewares']:
                if middleware.__class__.__name__ == 'RedisMiddleware':
                    self.ws_redis_connection = middleware.redis_connection_for_ws()
                    break
            else:
                self.ws_redis_connection = None

            # Don't create Manager() if we are going to use Redis for PubSub
            manager = None if self.ws_redis_connection else Manager()
            config['websocket_connections'] = WebsocketConnections(manager=manager)

    async def __call__(self, scope: dict, receive: Callable, send: Callable) -> None:
        """
        1.
            await func(scope, receive, send)
        2.
            async with asyncio.TaskGroup() as tg:
                tg.create_task(func(scope, receive, send))
        3.
            async with anyio.create_task_group() as task_group:
                task_group.start_soon(func, scope, receive, send)
                await anyio.to_thread.run_sync(func, scope, receive, send)
        4.
            with ProcessPoolExecutor() as e:
                e.submit(func, scope, receive, send)
        """
        if scope['type'] == 'lifespan':
            message = await receive()
            if message["type"] == "lifespan.startup":
                await self.handle_ws_listener()
                await self.handle_startup()
            return

        func = self.handle_http if scope['type'] == 'http' else self.handle_ws
        await func(scope=scope, receive=receive, send=send)

    async def handle_ws(self, scope: dict, receive: Callable, send: Callable) -> None:
        from panther.websocket import GenericWebsocket, Websocket

        # Monitoring
        monitoring = Monitoring(is_active=config['monitoring'], is_ws=True)

        # Create Temp Connection
        temp_connection = Websocket(scope=scope, receive=receive, send=send)
        await monitoring.before(request=temp_connection)

        # Find Endpoint
        endpoint, found_path = find_endpoint(path=temp_connection.path)
        if endpoint is None:
            await monitoring.after('Rejected')
            return await temp_connection.close(status.WS_1000_NORMAL_CLOSURE)

        # Check Endpoint Type
        if not issubclass(endpoint, GenericWebsocket):
            logger.critical(f'You may have forgotten to inherit from GenericWebsocket on the {endpoint.__name__}()')
            await monitoring.after('Rejected')
            return await temp_connection.close(status.WS_1014_BAD_GATEWAY)

        # Collect Path Variables
        path_variables: dict = collect_path_variables(request_path=temp_connection.path, found_path=found_path)

        # Create The Connection
        del temp_connection
        connection = endpoint(scope=scope, receive=receive, send=send)
        connection.set_path_variables(path_variables=path_variables)

        # Call 'Before' Middlewares
        if await self._run_ws_middlewares_before_listen(connection=connection):
            # Only Listen() If Middlewares Didn't Raise Anything
            await config['websocket_connections'].new_connection(connection=connection)
            await monitoring.after('Accepted')
            await connection.listen()

        # Call 'After' Middlewares
        await self._run_ws_middlewares_after_listen(connection=connection)

        # Done
        await monitoring.after('Closed')
        return None

    @classmethod
    async def _run_ws_middlewares_before_listen(cls, *, connection) -> bool:
        for middleware in config['ws_middlewares']:
            try:
                connection = await middleware.before(request=connection)
            except APIException:
                await connection.close()
                return False
        return True

    @classmethod
    async def _run_ws_middlewares_after_listen(cls, *, connection):
        for middleware in config['reversed_ws_middlewares']:
            with contextlib.suppress(APIException):
                await middleware.after(response=connection)

    async def handle_http(self, scope: dict, receive: Callable, send: Callable) -> None:
        request = Request(scope=scope, receive=receive, send=send)

        # Monitoring
        await self.monitoring.before(request=request)

        # Read Request Payload
        await request.read_body()

        # Find Endpoint
        _endpoint, found_path = find_endpoint(path=request.path)
        if _endpoint is None:
            return await self._raise(send, status_code=status.HTTP_404_NOT_FOUND)

        # Check Endpoint Type
        try:
            if isinstance(_endpoint, types.FunctionType):
                endpoint = check_function_type_endpoint(endpoint=_endpoint)
            else:
                endpoint = check_class_type_endpoint(endpoint=_endpoint)
        except TypeError:
            return await self._raise(send, status_code=status.HTTP_501_NOT_IMPLEMENTED)

        # Collect Path Variables
        path_variables: dict = collect_path_variables(request_path=request.path, found_path=found_path)

        try:  # They Both(middleware.before() & _endpoint()) Have The Same Exception (APIException)
            # Call 'Before' Middlewares
            for middleware in config['http_middlewares']:
                request = await middleware.before(request=request)

            # Call Endpoint
            response = await endpoint(request=request, **path_variables)

        except APIException as e:
            response = self._handle_exceptions(e)

        except Exception as e:  # noqa: BLE001
            # Every unhandled exception in Panther or code will catch here
            exception = clean_traceback_message(exception=e)
            logger.critical(exception)
            return await self._raise(send, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        # Call 'After' Middleware
        for middleware in config['reversed_http_middlewares']:
            try:
                response = await middleware.after(response=response)
            except APIException as e:  # noqa: PERF203
                response = self._handle_exceptions(e)

        await http_response(
            send,
            status_code=response.status_code,
            monitoring=self.monitoring,
            headers=response.headers,
            body=response.body,
        )

    async def handle_ws_listener(self):
        # Start Websocket Listener (Redis/ Queue)
        if config['has_ws']:
            Thread(
                target=config['websocket_connections'],
                daemon=True,
                args=(self.ws_redis_connection,),
            ).start()

    async def handle_startup(self):
        if startup := config['startup'] or self._startup:
            if is_function_async(startup):
                await startup()
            else:
                startup()

    def handle_shutdown(self):
        if shutdown := config['shutdown'] or self._shutdown:
            if is_function_async(shutdown):
                try:
                    asyncio.run(shutdown())
                except ModuleNotFoundError:
                    # Error: import of asyncio halted; None in sys.modules
                    #   And as I figured it out, it only happens when we running with
                    #   gunicorn and Uvicorn workers (-k uvicorn.workers.UvicornWorker)
                    pass
            else:
                shutdown()

    def __del__(self):
        self.handle_shutdown()

    @classmethod
    def _handle_exceptions(cls, e: APIException, /) -> Response:
        return Response(
            data=e.detail if isinstance(e.detail, dict) else {'detail': e.detail},
            status_code=e.status_code,
        )

    async def _raise(self, send, *, status_code: int):
        await http_response(
            send,
            headers={'content-type': 'application/json'},
            status_code=status_code,
            monitoring=self.monitoring,
            exception=True,
        )



from collections import namedtuple
from datetime import timedelta, datetime
import logging
from types import NoneType

import orjson as json

from panther.configs import config
from panther.db.connection import redis
from panther.request import Request
from panther.response import Response, ResponseDataTypes
from panther.utils import generate_hash_value_from_string, round_datetime

logger = logging.getLogger('panther')

caches = {}
CachedResponse = namedtuple('Cached', ['data', 'status_code'])


def cache_key(request: Request, /) -> str:
    client = request.user and request.user.id or request.client.ip
    query_params_hash = generate_hash_value_from_string(request.scope['query_string'].decode('utf-8'))
    return f'{client}-{request.path}-{query_params_hash}-{request.validated_data}'


def local_cache_key(*, request: Request, cache_exp_time: timedelta | None = None) -> str:
    key = cache_key(request)
    if cache_exp_time:
        time = round_datetime(datetime.now(), cache_exp_time)
        return f'{time}-{key}'
    else:
        return key


def get_cached_response_data(*, request: Request, cache_exp_time: timedelta) -> CachedResponse | None:
    """
    If redis.is_connected:
        Get Cached Data From Redis
    else:
        Get Cached Data From Memory
    """
    if redis.is_connected:  # noqa: Unresolved References
        key = cache_key(request)
        data = (redis.get(key) or b'{}').decode()
        if cached_value := json.loads(data):
            return CachedResponse(*cached_value)

    else:
        key = local_cache_key(request=request, cache_exp_time=cache_exp_time)
        if cached_value := caches.get(key):
            return CachedResponse(*cached_value)

    return None


def set_cache_response(*, request: Request, response: Response, cache_exp_time: timedelta | int) -> None:
    """
    If redis.is_connected:
        Cache The Data In Redis
    else:
        Cache The Data In Memory
    """

    cache_data: tuple[ResponseDataTypes, int] = (response.data, response.status_code)

    if redis.is_connected:  # noqa: Unresolved References
        key = cache_key(request)

        cache_exp_time = cache_exp_time or config['default_cache_exp']
        cache_data: bytes = json.dumps(cache_data)

        if not isinstance(cache_exp_time, timedelta | int | NoneType):
            msg = '"cache_exp_time" should be instance of "datetime.timedelta" or "int" or "None"'
            raise TypeError(msg)

        if cache_exp_time is None:
            logger.warning(
                'your response are going to cache in redis forever '
                '** set DEFAULT_CACHE_EXP in configs or pass the cache_exp_time in @API.get() for prevent this **'
            )
            redis.set(key, cache_data)
        else:
            redis.set(key, cache_data, ex=cache_exp_time)

    else:
        key = local_cache_key(request=request, cache_exp_time=cache_exp_time)
        caches[key] = cache_data

        if cache_exp_time:
            logger.info('"cache_exp_time" is not very accurate when redis is not connected.')



from pydantic import create_model
from pydantic_core._pydantic_core import PydanticUndefined


class ModelSerializer:
    def __new__(cls, *args, model=None, **kwargs):
        # Check `metaclass`
        if len(args) == 0:
            address = f'{cls.__module__}.{cls.__name__}'
            msg = f"you should not inherit the 'ModelSerializer', you should use it as 'metaclass' -> {address}"
            raise TypeError(msg)

        model_name = args[0]
        data = args[2]
        address = f'{data["__module__"]}.{model_name}'

        # Check `model`
        if model is None:
            msg = f"'model' required while using 'ModelSerializer' metaclass -> {address}"
            raise AttributeError(msg)
        # Check `fields`
        if 'fields' not in data:
            msg = f"'fields' required while using 'ModelSerializer' metaclass. -> {address}"
            raise AttributeError(msg) from None

        model_fields = model.model_fields
        field_definitions = {}

        # Collect `fields`
        for field_name in data['fields']:
            if field_name not in model_fields:
                msg = f"'{field_name}' is not in '{model.__name__}' -> {address}"
                raise AttributeError(msg) from None
            field_definitions[field_name] = (model_fields[field_name].annotation, model_fields[field_name])

        # Change `required_fields
        for required in data.get('required_fields', []):
            if required not in field_definitions:
                msg = f"'{required}' is in 'required_fields' but not in 'fields' -> {address}"
                raise AttributeError(msg) from None
            field_definitions[required][1].default = PydanticUndefined

        # Create Model
        return create_model(
            __model_name=model_name,
            **field_definitions
        )



import asyncio
import datetime
import logging
import sys
import time
from threading import Thread
from typing import Callable, Literal

from panther._utils import is_function_async
from panther.utils import Singleton


__all__ = (
    'BackgroundTask',
    'background_tasks',
)


logger = logging.getLogger('panther')


if sys.version_info.minor >= 11:
    from typing import Self
else:
    from typing import TypeVar

    Self = TypeVar('Self', bound='BackgroundTask')


class BackgroundTask:
    """
    Default: Task is going to run once,
    and if you only specify a custom interval, default interval time is 1 minutes
    """
    def __init__(self, func, *args, **kwargs):
        self._func: Callable = func
        self._args: tuple = args
        self._kwargs: dict = kwargs
        self._remaining_interval: int = 1
        self._last_run: datetime.datetime | None = None
        self._timedelta: datetime.timedelta = datetime.timedelta(minutes=1)
        self._time: datetime.time | None = None
        self._unit: Literal['seconds', 'minutes', 'hours', 'days', 'weeks'] | None = None

    def interval(self, interval: int, /) -> Self:
        """
        interval = -1 --> Infinite
        """
        self._remaining_interval = interval
        return self

    def every_seconds(self, seconds: int = 1, /) -> Self:
        """
        Every How Many Seconds? (Default is 1)
        """
        self._unit = 'seconds'
        self._timedelta = datetime.timedelta(seconds=seconds)
        return self

    def every_minutes(self, minutes: int = 1, /) -> Self:
        """
        Every How Many Minutes? (Default is 1)
        """
        self._unit = 'minutes'
        self._timedelta = datetime.timedelta(minutes=minutes)
        return self

    def every_hours(self, hours: int = 1, /) -> Self:
        """
        Every How Many Hours? (Default is 1)
        """
        self._unit = 'hours'
        self._timedelta = datetime.timedelta(hours=hours)
        return self

    def every_days(self, days: int = 1, /) -> Self:
        """
        Every How Many Days? (Default is 1)
        """
        self._unit = 'days'
        self._timedelta = datetime.timedelta(days=days)
        return self

    def every_weeks(self, weeks: int = 1, /) -> Self:
        """
        Every How Many Weeks? (Default is 1)
        """
        self._unit = 'weeks'
        self._timedelta = datetime.timedelta(weeks=weeks)
        return self

    # TODO: Coming Soon
    # def on(
    #         self,
    #         day_of_week: Literal['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday'],
    #         /
    # ) -> Self:
    #     """
    #     Set day to schedule the task, useful on `.every_weeks()`
    #     """
    #     if self._unit != 'weeks':
    #         logger.warning('`.on()` only useful when you are using `.every_weeks()`')
    #     return self

    def at(self, _time: datetime.time, /) -> Self:
        """
        Set a time to schedule the task,
        Only useful on `.every_days()` and `.every_weeks()`
        """
        if isinstance(_time, datetime.time):
            self._time = _time
        elif isinstance(_time, datetime.datetime):
            _time = _time.time()
        else:
            raise TypeError(
                f'Argument should be instance of `datetime.time()` or `datetime.datetime()` not `{type(_time)}`')

        if self._unit not in ['days', 'weeks']:
            logger.warning('`.at()` only useful when you are using `.every_days()` or `.every_weeks()`')

        return self

    def _should_wait(self) -> bool:
        """
        Return:
        ------
            True: Wait and do nothing
            False: Don't wait and Run this task
        """
        now = datetime.datetime.now()

        # Wait
        if self._last_run and (self._last_run + self._timedelta) > now:
            return True

        # We don't have time condition, so run
        if self._time is None:
            self._last_run = now
            return False

        # Time is ok, so run
        if bool(
                now.hour == self._time.hour and
                now.minute == self._time.minute and
                now.second == self._time.second,
        ):
            self._last_run = now
            return False

        # Time was not ok
        return True

    def __call__(self) -> bool:
        """
        Return:
        ------
            True: Everything is ok
            False: This task is done, remove it from `BackgroundTasks.tasks`
        """
        if self._remaining_interval == 0:
            return False

        if self._should_wait():
            # Just wait, it's not your time yet :)
            return True

        logger.info(
            f'{self._func.__name__}('
            f'{", ".join(str(a) for a in self._args)}, '
            f'{", ".join(f"{k}={v}" for k, v in self._kwargs.items())}'
            f') Remaining Interval -> {"∞" if self._remaining_interval == -1 else self._remaining_interval - 1}',
        )
        if self._remaining_interval != -1:
            self._remaining_interval -= 1
        if is_function_async(self._func):
            asyncio.run(self._func(*self._args, **self._kwargs))
        else:
            self._func(*self._args, **self._kwargs)

        return True


class BackgroundTasks(Singleton):
    _initialized: bool = False

    def __init__(self):
        self.tasks = []

    def add_task(self, task: BackgroundTask):
        if self._initialized is False:
            logger.error('Task will be ignored, `BACKGROUND_TASKS` is not True in `core/configs.py`')
            return

        if not self._is_instance_of_task(task):
            return

        if task not in self.tasks:
            self.tasks.append(task)

    def initialize(self):
        """
        We only call initialize() once in the panther.main.Panther.load_configs()
        """

        def __run_task(task):
            if task() is False:
                self.tasks.remove(task)

        def __run_tasks():
            while True:
                [Thread(target=__run_task, args=(task,)).start() for task in self.tasks[:]]
                time.sleep(1)

        if self._initialized is False:
            self._initialized = True
            Thread(target=__run_tasks, daemon=True).start()

    @classmethod
    def _is_instance_of_task(cls, task, /):
        if not isinstance(task, BackgroundTask):
            name = getattr(task, '__name__', task.__class__.__name__)
            logger.error(f'`{name}` should be instance of `background_tasks.BackgroundTask`')
            return False
        return True


background_tasks = BackgroundTasks()

"""
-------------------------------------------------------------
Example:
-------------------------------------------------------------
>>> import datetime


>>> async def hello(name: str):
>>>     print(f'Hello {name}')

# Run it every 5 seconds for 2 times

>>> task1 = BackgroundTask(hello, 'Ali').interval(2).every_seconds(5)
>>> background_tasks.add_task(task1)

# Run it every day at 08:00 O'clock forever

>>> task2 = BackgroundTask(hello, 'Saba').interval(-1).every_days().at(datetime.time(hour=8))
>>> background_tasks.add_task(task2)
"""



from __future__ import annotations

import asyncio
import contextlib
import logging
from multiprocessing import Manager
from typing import TYPE_CHECKING, Literal

import orjson as json

from panther import status
from panther._utils import generate_ws_connection_id
from panther.base_request import BaseRequest
from panther.configs import config
from panther.db.connection import redis
from panther.utils import Singleton

if TYPE_CHECKING:
    from redis import Redis

logger = logging.getLogger('panther')


class PubSub:
    def __init__(self, manager):
        self._manager = manager
        self._subscribers = self._manager.list()

    def subscribe(self):
        queue = self._manager.Queue()
        self._subscribers.append(queue)
        return queue

    def publish(self, msg):
        for queue in self._subscribers:
            queue.put(msg)


class WebsocketConnections(Singleton):
    def __init__(self, manager: Manager = None):
        self.connections = {}
        self.connections_count = 0
        self.manager = manager

    def __call__(self, r: Redis | None):
        if r:
            subscriber = r.pubsub()
            subscriber.subscribe('websocket_connections')
            logger.info("Subscribed to 'websocket_connections' channel")
            for channel_data in subscriber.listen():
                match channel_data['type']:
                    # Subscribed
                    case 'subscribe':
                        continue

                    # Message Received
                    case 'message':
                        loaded_data = json.loads(channel_data['data'].decode())
                        self._handle_received_message(received_message=loaded_data)

                    case unknown_type:
                        logger.debug(f'Unknown Channel Type: {unknown_type}')
        else:
            self.pubsub = PubSub(manager=self.manager)
            queue = self.pubsub.subscribe()
            logger.info("Subscribed to 'websocket_connections' queue")
            while True:
                received_message = queue.get()
                self._handle_received_message(received_message=received_message)

    def _handle_received_message(self, received_message):
        if (
                isinstance(received_message, dict)
                and (connection_id := received_message.get('connection_id'))
                and connection_id in self.connections
                and 'action' in received_message
                and 'data' in received_message
        ):
            # Check Action of WS
            match received_message['action']:
                case 'send':
                    asyncio.run(self.connections[connection_id].send(data=received_message['data']))
                case 'close':
                    with contextlib.suppress(RuntimeError):
                        asyncio.run(self.connections[connection_id].close(
                            code=received_message['data']['code'],
                            reason=received_message['data']['reason']
                        ))
                        # We are trying to disconnect the connection between a thread and a user
                        # from another thread, it's working, but we have to find another solution for it
                        #
                        # Error:
                        # Task <Task pending coro=<Websocket.close()>> got Future
                        # <Task pending coro=<WebSocketCommonProtocol.transfer_data()>>
                        # attached to a different loop
                case unknown_action:
                    logger.debug(f'Unknown Message Action: {unknown_action}')

    def publish(self, connection_id: str, action: Literal['send', 'close'], data: any):
        publish_data = {'connection_id': connection_id, 'action': action, 'data': data}

        if redis.is_connected:
            redis.publish('websocket_connections', json.dumps(publish_data))
        else:
            self.pubsub.publish(publish_data)

    async def new_connection(self, connection: Websocket) -> None:
        await connection.connect(**connection.path_variables)
        if not hasattr(connection, '_connection_id'):
            # User didn't even call the `self.accept()` so close the connection
            await connection.close()

        if connection.is_connected:
            self.connections_count += 1

            # Save New ConnectionID
            self.connections[connection.connection_id] = connection

    def remove_connection(self, connection: Websocket) -> None:
        if connection.is_connected:
            self.connections_count -= 1
            del self.connections[connection.connection_id]


class Websocket(BaseRequest):
    is_connected: bool = False

    async def connect(self, **kwargs) -> None:
        """Check your conditions then self.accept() the connection"""
        await self.accept()

    async def accept(self, subprotocol: str | None = None, headers: dict | None = None) -> None:
        await self.asgi_send({'type': 'websocket.accept', 'subprotocol': subprotocol, 'headers': headers or {}})
        self.is_connected = True

        # Generate ConnectionID
        connection_id = generate_ws_connection_id()
        while connection_id in config['websocket_connections'].connections:
            connection_id = generate_ws_connection_id()

        # Set ConnectionID
        self.set_connection_id(connection_id)

    async def receive(self, data: str | bytes) -> None:
        pass

    async def send(self, data: any = None) -> None:
        logger.debug(f'Sending WS Message to {self.connection_id}')
        if data:
            if isinstance(data, bytes):
                await self.send_bytes(bytes_data=data)
            elif isinstance(data, str):
                await self.send_text(text_data=data)
            else:
                await self.send_text(text_data=json.dumps(data).decode())

    async def send_text(self, text_data: str) -> None:
        await self.asgi_send({'type': 'websocket.send', 'text': text_data})

    async def send_bytes(self, bytes_data: bytes) -> None:
        await self.asgi_send({'type': 'websocket.send', 'bytes': bytes_data})

    async def close(self, code: int = status.WS_1000_NORMAL_CLOSURE, reason: str = '') -> None:
        logger.debug(f'Closing WS Connection {self.connection_id}')
        self.is_connected = False
        config['websocket_connections'].remove_connection(self)
        await self.asgi_send({'type': 'websocket.close', 'code': code, 'reason': reason})

    async def listen(self) -> None:
        while self.is_connected:
            response = await self.asgi_receive()
            if response['type'] == 'websocket.connect':
                continue

            if response['type'] == 'websocket.disconnect':
                break

            if 'text' in response:
                await self.receive(data=response['text'])
            else:
                await self.receive(data=response['bytes'])

    def set_path_variables(self, path_variables: dict) -> None:
        self._path_variables = path_variables

    @property
    def path_variables(self) -> dict:
        return getattr(self, '_path_variables', {})

    def set_connection_id(self, connection_id: str) -> None:
        self._connection_id = connection_id

    @property
    def connection_id(self) -> str:
        connection_id = getattr(self, '_connection_id', None)
        if connection_id is None:
            logger.error('You should first `self.accept()` the connection then use the `self.connection_id`')
        return connection_id



import logging
from time import perf_counter
from typing import Literal

from panther.base_request import BaseRequest


logger = logging.getLogger('monitoring')


class Monitoring:
    """
    Create Log Message Like Below:
    date time | method | path | ip:port | response_time [ms, s] | status
    """
    def __init__(self, is_active: bool, is_ws: bool = False):
        self.is_active = is_active
        self.is_ws = is_ws

    async def before(self, request: BaseRequest):
        if self.is_active:
            ip, port = request.client

            if self.is_ws:
                method = 'WS'
            else:
                method = request.scope['method']

            self.log = f'{method} | {request.path} | {ip}:{port}'
            self.start_time = perf_counter()

    async def after(self, status: int | Literal['Accepted', 'Rejected', 'Closed'], /):
        if self.is_active:
            response_time = perf_counter() - self.start_time
            time_unit = ' s'

            if response_time < 0.01:
                response_time = response_time * 1_000
                time_unit = 'ms'

            logger.info(f'{self.log} | {round(response_time, 4)} {time_unit} | {status}')



from pathlib import Path
from unittest import TestCase

from pydantic import Field

from panther import Panther
from panther.app import API
from panther.db import Model
from panther.request import Request
from panther.serializer import ModelSerializer
from panther.test import APIClient


class Book(Model):
    name: str
    author: str = Field('default_author')
    pages_count: int = Field(0)


class NotRequiredFieldsSerializer(metaclass=ModelSerializer, model=Book):
    fields = ['author', 'pages_count']


class RequiredFieldsSerializer(metaclass=ModelSerializer, model=Book):
    fields = ['name', 'author', 'pages_count']


class OnlyRequiredFieldsSerializer(metaclass=ModelSerializer, model=Book):
    fields = ['name', 'author', 'pages_count']
    required_fields = ['author', 'pages_count']


@API(input_model=NotRequiredFieldsSerializer)
async def not_required(request: Request):
    return request.validated_data


@API(input_model=RequiredFieldsSerializer)
async def required(request: Request):
    return request.validated_data


@API(input_model=OnlyRequiredFieldsSerializer)
async def only_required(request: Request):
    return request.validated_data


urls = {
    'not-required': not_required,
    'required': required,
    'only-required': only_required,
}


class TestModelSerializer(TestCase):
    DB_PATH = 'test.pdb'

    @classmethod
    def setUpClass(cls) -> None:
        global MIDDLEWARES
        MIDDLEWARES = [
            ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{cls.DB_PATH}'}),
        ]
        app = Panther(__name__, configs=__name__, urls=urls)
        cls.client = APIClient(app=app)

    def tearDown(self) -> None:
        Path(self.DB_PATH).unlink(missing_ok=True)

    def test_not_required_fields_empty_response(self):
        payload = {}
        res = self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'default_author', 'pages_count': 0}

    def test_not_required_fields_full_response(self):
        payload = {
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('not-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'author': 'ali', 'pages_count': 12}

    def test_required_fields_error(self):
        payload = {}
        res = self.client.post('required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required'}

    def test_required_fields_success(self):
        payload = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    def test_only_required_fields_error(self):
        payload = {}
        res = self.client.post('only-required', payload=payload)
        assert res.status_code == 400
        assert res.data == {'name': 'Field required', 'author': 'Field required', 'pages_count': 'Field required'}

    def test_only_required_fields_success(self):
        payload = {
            'name': 'how to code',
            'author': 'ali',
            'pages_count': '12'
        }
        res = self.client.post('only-required', payload=payload)
        assert res.status_code == 200
        assert res.data == {'name': 'how to code', 'author': 'ali', 'pages_count': 12}

    def test_define_class_without_fields(self):
        try:
            class Serializer1(metaclass=ModelSerializer, model=Book):
                pass
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == "'fields' required while using 'ModelSerializer' metaclass. -> tests.test_model_serializer.Serializer1"
        else:
            assert False

    def test_define_class_with_invalid_fields(self):
        try:
            class Serializer2(metaclass=ModelSerializer, model=Book):
                fields = ['ok', 'no']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == "'ok' is not in 'Book' -> tests.test_model_serializer.Serializer2"
        else:
            assert False

    def test_define_class_with_invalid_required_fields(self):
        try:
            class Serializer3(metaclass=ModelSerializer, model=Book):
                fields = ['name', 'author']
                required_fields = ['pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == "'pages_count' is in 'required_fields' but not in 'fields' -> tests.test_model_serializer.Serializer3"
        else:
            assert False

    def test_define_class_without_model(self):
        try:
            class Serializer4(metaclass=ModelSerializer):
                fields = ['name', 'author']
                required_fields = ['pages_count']
        except Exception as e:
            assert isinstance(e, AttributeError)
            assert e.args[0] == "'model' required while using 'ModelSerializer' metaclass -> tests.test_model_serializer.Serializer4"
        else:
            assert False

    def test_define_class_without_metaclass(self):
        class Serializer5(ModelSerializer):
            fields = ['name', 'author']
            required_fields = ['pages_count']

        try:
            Serializer5(name='alice', author='bob')
        except Exception as e:
            assert isinstance(e, TypeError)
            assert e.args[0] == "you should not inherit the 'ModelSerializer', you should use it as 'metaclass' -> tests.test_model_serializer.Serializer5"
        else:
            assert False



import functools
from collections.abc import Callable
from datetime import datetime, timedelta
import logging
from typing import Literal

from orjson import JSONDecodeError
from pydantic import ValidationError

from panther import status
from panther._utils import is_function_async
from panther.caching import cache_key, get_cached_response_data, set_cache_response
from panther.configs import config
from panther.exceptions import (
    APIException,
    AuthorizationException,
    InvalidPathVariableException,
    JsonDecodeException,
    MethodNotAllowed,
    ThrottlingException,
)
from panther.request import Request
from panther.response import Response
from panther.throttling import Throttling, throttling_storage
from panther.utils import round_datetime

__all__ = ('API', 'GenericAPI')


logger = logging.getLogger('panther')


class API:
    def __init__(
            self,
            *,
            input_model=None,
            output_model=None,
            auth: bool = False,
            permissions: list | None = None,
            throttling: Throttling = None,
            cache: bool = False,
            cache_exp_time: timedelta | int | None = None,
            methods: list[Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']] | None = None,
    ):
        self.input_model = input_model
        self.output_model = output_model
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.cache_exp_time = cache_exp_time
        self.methods = methods
        self.request: Request | None = None

    def __call__(self, func):
        @functools.wraps(func)
        async def wrapper(request: Request, **path_variables) -> Response:
            self.request: Request = request  # noqa: Non-self attribute could not be type hinted

            # 1. Check Method
            if self.methods and self.request.method not in self.methods:
                raise MethodNotAllowed

            # 2. Authentication
            self.handle_authentications()

            # 3. Throttling
            self.handle_throttling()

            # 4. Permissions
            self.handle_permissions()

            # 5. Validate Input
            if self.request.method in ['POST', 'PUT', 'PATCH']:
                self.handle_input_validation()

            # 6. Get Cached Response
            if self.cache and self.request.method == 'GET':
                if cached := get_cached_response_data(request=self.request,  cache_exp_time=self.cache_exp_time):
                    return Response(data=cached.data, status_code=cached.status_code)

            # 7. Clean Path Variables
            self.clean_path_variables(func, path_variables)

            # 8. Put Request In kwargs (If User Wants It)
            kwargs = {}
            if req_arg := [k for k, v in func.__annotations__.items() if v == Request]:
                kwargs[req_arg[0]] = self.request

            # 9. Call Endpoint
            if is_function_async(func):
                response = await func(**kwargs, **path_variables)
            else:
                response = func(**kwargs, **path_variables)

            # 10. Clean Response
            if not isinstance(response, Response):
                response = Response(data=response)
            response._clean_data_with_output_model(output_model=self.output_model)  # noqa: SLF001

            # 11. Set New Response To Cache
            if self.cache and self.request.method == 'GET':
                set_cache_response(
                    request=self.request,
                    response=response,
                    cache_exp_time=self.cache_exp_time
                )

            # 12. Warning CacheExpTime
            if self.cache_exp_time and self.cache is False:
                logger.warning('"cache_exp_time" won\'t work while "cache" is False')

            return response

        return wrapper

    def handle_authentications(self) -> None:
        auth_class = config['authentication']
        if self.auth:
            if not auth_class:
                logger.critical('"AUTHENTICATION" has not been set in configs')
                raise APIException
            user = auth_class.authentication(self.request)
            self.request.set_user(user=user)

    def handle_throttling(self) -> None:
        if throttling := self.throttling or config['throttling']:
            key = cache_key(self.request)
            time = round_datetime(datetime.now(), throttling.duration)
            throttling_key = f'{time}-{key}'
            if throttling_storage[throttling_key] > throttling.rate:
                raise ThrottlingException

            throttling_storage[throttling_key] += 1

    def handle_permissions(self) -> None:
        for perm in self.permissions:
            if type(perm.authorization).__name__ != 'method':
                logger.error(f'{perm.__name__}.authorization should be "classmethod"')
                continue
            if perm.authorization(request=self.request) is False:
                raise AuthorizationException

    def handle_input_validation(self):
        if self.input_model:
            validated_data = self.validate_input(model=self.input_model, request=self.request)
            self.request.set_validated_data(validated_data)

    @classmethod
    def validate_input(cls, model, request: Request):
        try:
            if isinstance(request.data, bytes):
                raise APIException(detail='Content-Type is not valid', status_code=status.HTTP_400_BAD_REQUEST)
            return model(**request.data)
        except ValidationError as validation_error:
            error = {'.'.join(loc for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise APIException(detail=error, status_code=status.HTTP_400_BAD_REQUEST)
        except JSONDecodeError:
            raise JsonDecodeException

    @staticmethod
    def clean_path_variables(func: Callable, request_path_variables: dict):
        for name, value in request_path_variables.items():
            for variable_name, variable_type in func.__annotations__.items():
                if name == variable_name:
                    # Check the type and convert the value
                    if variable_type is bool:
                        request_path_variables[name] = value.lower() not in ['false', '0']

                    elif variable_type is int:
                        try:
                            request_path_variables[name] = int(value)
                        except ValueError:
                            raise InvalidPathVariableException(value=value, variable_type=variable_type)
                    break


class GenericAPI:
    input_model = None
    output_model = None
    auth: bool = False
    permissions: list | None = None
    throttling: Throttling | None = None
    cache: bool = False
    cache_exp_time: timedelta | int | None = None

    async def get(self, *args, **kwargs):
        raise MethodNotAllowed

    async def post(self, *args, **kwargs):
        raise MethodNotAllowed

    async def put(self, *args, **kwargs):
        raise MethodNotAllowed

    async def patch(self, *args, **kwargs):
        raise MethodNotAllowed

    async def delete(self, *args, **kwargs):
        raise MethodNotAllowed

    @classmethod
    async def call_method(cls, *args, **kwargs):
        match kwargs['request'].method:
            case 'GET':
                func = cls().get
            case 'POST':
                func = cls().post
            case 'PUT':
                func = cls().put
            case 'PATCH':
                func = cls().patch
            case 'DELETE':
                func = cls().delete

        return await API(
            input_model=cls.input_model,
            output_model=cls.output_model,
            auth=cls.auth,
            permissions=cls.permissions,
            throttling=cls.throttling,
            cache=cls.cache,
            cache_exp_time=cls.cache_exp_time,
        )(func)(*args, **kwargs)



from pathlib import Path
from unittest import TestCase

from panther import Panther
from panther.test import APIClient

DB_PATH = 'test.pdb'
MIDDLEWARES = [
    ('panther.middlewares.db.DatabaseMiddleware', {'url': f'pantherdb://{DB_PATH}'}),
]


class TestPanelAPIs(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        app = Panther(__name__, configs=__name__, urls={})
        cls.client = APIClient(app=app)

    @classmethod
    def tearDownClass(cls) -> None:
        Path(DB_PATH).unlink()

    def test_list_of_models(self):
        response = self.client.get('_panel')
        expected_keys = ['name', 'module', 'index']
        assert expected_keys == [*response.data[0].keys()]



import logging
import time
from abc import abstractmethod
from typing import Literal

from panther.cli.utils import import_error
from panther.configs import config
from panther.db.models import BaseUser, Model
from panther.exceptions import AuthenticationException
from panther.request import Request

try:
    from jose import JWTError, jwt
except ModuleNotFoundError as e:
    import_error(e, package='python-jose')

logger = logging.getLogger('panther')


class BaseAuthentication:
    @classmethod
    @abstractmethod
    def authentication(cls, request: Request):
        """Return Instance of User"""
        msg = f'{cls.__name__}.authentication() is not implemented.'
        raise cls.exception(msg) from None

    @staticmethod
    def exception(message: str, /) -> type[AuthenticationException]:
        logger.error(f'Authentication Error: "{message}"')
        return AuthenticationException


class JWTAuthentication(BaseAuthentication):
    model = BaseUser
    keyword = 'Bearer'
    algorithm = 'HS256'
    HTTP_HEADER_ENCODING = 'iso-8859-1'  # Header encoding (see RFC5987)

    @classmethod
    def get_authorization_header(cls, request: Request) -> bytes:
        auth = request.headers.authorization

        if auth is None:
            msg = 'Authorization is required'
            raise cls.exception(msg) from None

        if isinstance(auth, str):
            try:
                auth = auth.encode(JWTAuthentication.HTTP_HEADER_ENCODING)
            except UnicodeEncodeError as e:
                raise cls.exception(e) from None

        return auth

    @classmethod
    def authentication(cls, request: Request) -> Model:
        auth = cls.get_authorization_header(request).split()
        if not auth or auth[0].lower() != cls.keyword.lower().encode():
            msg = 'Authorization keyword is not valid'
            raise cls.exception(msg) from None
        if len(auth) != 2:
            msg = 'Authorization should have 2 part'
            raise cls.exception(msg) from None

        try:
            token = auth[1].decode()
        except UnicodeError:
            msg = 'Unicode Error'
            raise cls.exception(msg) from None

        payload = cls.decode_jwt(token)
        return cls.get_user(payload)

    @classmethod
    def get_user(cls, payload: dict) -> Model:
        """Get UserModel from config, else use default UserModel from cls.model"""
        if (user_id := payload.get('user_id')) is None:
            msg = 'Payload does not have user_id'
            raise cls.exception(msg)

        user_model = config['user_model'] or cls.model
        if user := user_model.find_one(id=user_id):
            return user

        msg = 'User not found'
        raise cls.exception(msg) from None

    @classmethod
    def encode_jwt(cls, user_id: str, token_type: Literal['access', 'refresh'] = 'access') -> str:
        """Encode JWT from user_id."""
        issued_at = time.time()
        if token_type == 'access':
            expire = issued_at + config['jwt_config'].life_time
        else:
            expire = issued_at + config['jwt_config'].refresh_life_time

        claims = {
            'token_type': token_type,
            'user_id': user_id,
            'iat': issued_at,
            'exp': expire,
        }
        return jwt.encode(
            claims,
            key=config['jwt_config'].key,
            algorithm=config['jwt_config'].algorithm,
        )

    @classmethod
    def encode_refresh_token(cls, user_id: str) -> str:
        """Encode JWT from user_id."""
        return cls.encode_jwt(user_id=user_id, token_type='refresh')

    @classmethod
    def decode_jwt(cls, token: str) -> dict:
        """Decode JWT token to user_id (it can return multiple variable ... )"""
        try:
            return jwt.decode(
                token=token,
                key=config['jwt_config'].key,
                algorithms=[config['jwt_config'].algorithm],
            )
        except JWTError as e:
            raise cls.exception(e) from None

    @classmethod
    def login(cls, user_id: str) -> str:
        """Alias of encode_jwt()"""
        return cls.encode_jwt(user_id=user_id)

    @staticmethod
    def exception(message: str | JWTError | UnicodeEncodeError, /) -> type[AuthenticationException]:
        logger.error(f'JWT Authentication Error: "{message}"')
        return AuthenticationException



import base64
import hashlib
import logging
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import ClassVar

logger = logging.getLogger('panther')

URANDOM_SIZE = 16


class Singleton(object):
    _instances: ClassVar[dict] = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls)
        return cls._instances[cls]


def load_env(env_file: str | Path, /) -> dict[str, str]:
    variables = {}

    if env_file is None or not Path(env_file).is_file():
        logger.critical(f'"{env_file}" is not valid file for load_env()')
        return variables

    with open(env_file) as file:
        for line in file.readlines():
            striped_line = line.strip()
            if not striped_line.startswith('#') and '=' in striped_line:
                key, value = striped_line.split('=', 1)
                key = key.strip()
                value = value.strip().strip('"\'')
                variables[key] = value

                # Load them as system environment variable
                os.environ[key] = value
    return variables


def generate_secret_key() -> str:
    return base64.urlsafe_b64encode(os.urandom(32)).decode()


def round_datetime(dt: datetime, delta: timedelta):
    return datetime.min + round((dt - datetime.min) / delta) * delta


def generate_hash_value_from_string(string_value: str, /) -> str:
    # The point of this method is for maintenance, if we want to change
    # the hash algorithm in the future, it will be easy.
    return hashlib.sha256(string_value.encode('utf-8')).hexdigest()


def scrypt(password: str, salt: bytes, digest: bool = False) -> str | bytes:
    """
    n: CPU/memory cost parameter – Must be a power of 2 (e.g. 1024)
    r: Block size parameter, which fine-tunes sequential memory read size and performance. (8 is commonly used)
    p: Parallelization parameter. (1 .. 232-1 * hLen/MFlen)
    dk_len: Desired key length in bytes (
        Intended output length in octets of the derived key; a positive integer satisfying dkLen ≤ (232− 1) * hLen.)
    h_len: The length in octets of the hash function (32 for SHA256).
    mf_len: The length in octets of the output of the mixing function (SMix below). Defined as r * 128 in RFC7914.
    """
    n = 2 ** 14  # 16384
    r = 8
    p = 10
    dk_len = 64

    derived_key = hashlib.scrypt(
        password=password.encode(),
        salt=salt,
        n=n,
        r=r,
        p=p,
        dklen=dk_len
    )
    if digest:

        return hashlib.md5(derived_key).hexdigest()
    else:
        return derived_key


def encrypt_password(password: str) -> str:
    salt = os.urandom(URANDOM_SIZE)
    derived_key = scrypt(password=password, salt=salt, digest=True)

    return f'{salt.hex()}{derived_key}'


def check_password(stored_password: str, new_password: str) -> bool:
    size = URANDOM_SIZE * 2
    salt = stored_password[:size]
    stored_hash = stored_password[size:]
    derived_key = scrypt(password=new_password, salt=bytes.fromhex(salt), digest=True)

    return derived_key == stored_hash



from datetime import datetime

import bson
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, field_validator

from panther.configs import config
from panther.db.queries import Query


class Model(PydanticBaseModel, Query):
    id: str | None = Field(None, validation_alias='_id')

    @field_validator('id', mode='before')
    def validate_id(cls, value: str | int | bson.ObjectId) -> str:
        if isinstance(value, int):
            pass

        elif isinstance(value, str):
            try:
                bson.ObjectId(value)
            except bson.objectid.InvalidId as e:
                msg = 'Invalid ObjectId'
                raise ValueError(msg) from e

        elif not isinstance(value, bson.ObjectId):
            msg = 'ObjectId required'
            raise ValueError(msg) from None

        return str(value)

    @property
    def _id(self) -> int | bson.ObjectId | None:
        if config['query_engine'].__name__ == 'BasePantherDBQuery':
            return int(self.id)
        else:
            return bson.ObjectId(self.id) if self.id else None

    def dict(self, *args, **kwargs) -> dict:
        return self.model_dump(*args, **kwargs)


class BaseUser(Model):
    first_name: str = Field('', max_length=64)
    last_name: str = Field('', max_length=64)
    last_login: datetime = None

    def update_last_login(self) -> None:
        self.update(last_login=datetime.now())



from panther import status
from panther.app import API
from panther.configs import config
from panther.panel.utils import get_model_fields
from panther.request import Request
from panther.response import Response


@API(methods=['GET'])
async def models_api():
    return [{
        'name': m['name'],
        'module': m['module'],
        'index': i
    } for i, m in enumerate(config['models'])]


@API(methods=['GET', 'POST'])
async def documents_api(request: Request, index: int):
    model = config['models'][index]['class']

    if request.method == 'POST':
        validated_data = API.validate_input(model=model, request=request)
        document = model.insert_one(**validated_data.model_dump(exclude=['id']))
        return Response(data=document, status_code=status.HTTP_201_CREATED)

    else:
        result = {
            'fields': get_model_fields(model),
        }
        if data := model.find():
            result['data'] = data
        else:
            result['data'] = []
        return result


@API(methods=['PUT', 'DELETE', 'GET'])
async def single_document_api(request: Request, index: int, document_id: int | str):
    model = config['models'][index]['class']

    if document := model.find_one(id=document_id):
        if request.method == 'PUT':
            validated_data = API.validate_input(model=model, request=request)
            document.update(**validated_data.model_dump(exclude=['id']))
            return Response(data=document, status_code=status.HTTP_202_ACCEPTED)

        elif request.method == 'DELETE':
            document.delete()
            return Response(status_code=status.HTTP_204_NO_CONTENT)

        else:  # GET
            return document

    else:
        return Response(status_code=status.HTTP_404_NOT_FOUND)



from panther.db.connection import DBSession
from panther.middlewares.base import BaseMiddleware
from panther.request import Request
from panther.response import Response
from panther.websocket import GenericWebsocket


class DatabaseMiddleware(BaseMiddleware):
    def __init__(self, **kwargs):
        self.url = kwargs['url']

    async def before(self, request: Request | GenericWebsocket) -> Request | GenericWebsocket:
        self.db = DBSession(db_url=self.url)
        return request

    async def after(self, response: Response | GenericWebsocket) -> Response | GenericWebsocket:
        self.db.close()
        return response
